from twisted.internet.defer import inlineCallbacks

@inlineCallbacks
def key_words(session, question=None, key_words=None, time=1000, debug=False):
    """
    This function asks a question and waits for the user to respond with a keyword from the list of keywords.

    Args:
        question (str): The question to be asked.
        key_words (list): A list of keywords to be checked in the user response.
        time (int): The time to wait for the user response in ms.
        debug (bool): A flag to print debug information.

    Returns:
        str: The keyword found in the user response.

    """
    # Check if the arguments are of the correct type
    if not isinstance(question, str):
        raise TypeError("question is not a string")
    if not isinstance(key_words, list):
        raise TypeError("key_words is not a list")
    # check if the list contains only strings
    else:
        for word in key_words:
            if not isinstance(word, str):
                raise TypeError("key_words is not a list of strings")
    if not isinstance(time, int):
        raise TypeError("time is not an integer")
    if not isinstance(debug, bool):
        raise TypeError("debug is not a boolean")

    # ask question
    yield session.call("rie.dialogue.say", text=question)
    # get user input and parse it
    user_input = yield session.call("rie.dialogue.stt.read", time=time)
    user_response = ""
    if debug:
        print("The entire user input is: ")
        print(user_input)

    for frame in user_input:
        if frame["data"]["body"]["final"]:
            user_response = frame["data"]["body"]["text"]

    if debug:
        print("The user response is : " + user_response)
    answer_found = None

    for word in user_response.split():
        word = word.lower()
        print(word)
        if word in key_words and answer_found is None:
            answer_found = word
            break
    if debug:
        print("The keyword found: " + str(answer_found))
    return answer_found



import cv2
import numpy as np
import base64


def aruco_detect_markers(frame):
    """
    Args:
    frame (dictionary):
        The frame dictionary from the robot's camera stream
    Returns
        corners: list
        The corners of the detected markers
        ids: list
        The ids of the detected markers
    """
    # check if the frame is not empty
    if frame is None:
        raise ValueError("The frame is empty")
    # check if the frame is a dictionary
    if not isinstance(frame, dict):
        raise TypeError("The frame is not a dictionary")

    frame_single = frame["data"]["body.head.eyes"]
    # make sure the frame is byte-like and not a string; it's in base64
    frame_single = bytes(frame_single, "utf-8")
    # Decode the base64 string
    image_data = base64.b64decode(frame_single)

    # Convert the decoded bytes to a numpy array
    nparr = np.frombuffer(image_data, np.uint8)

    # Decode the numpy array into an image
    image = cv2.imdecode(nparr, cv2.IMREAD_COLOR)

    # Load the dictionary that was used to generate the markers
    dictionary = cv2.aruco.getPredefinedDictionary(cv2.aruco.DICT_6X6_250)
    parameters = cv2.aruco.DetectorParameters()
    detector = cv2.aruco.ArucoDetector(dictionary, parameters)

    # Detect the markers in the image
    corners, ids, rejectedImgPoints = detector.detectMarkers(image)

    return corners, ids



from random import randint
from twisted.internet.defer import inlineCallbacks
from autobahn.twisted.util import sleep

user_response = ""
answers_found = False


def listen_smart_question(frames):
    global user_response
    if frames["data"]["body"]["final"]:
        print(frames["data"]["body"]["text"])
        user_response = frames["data"]["body"]["text"]


def find_the_answer(answer_dictionary):
    global answers_found
    answer = None
    for key in answer_dictionary.keys():  # It needs to search all values of the dictioinary, so all lists of strings and return the key
        for value in answer_dictionary[key]:
            if value in user_response:
                print("found the answer")
                answers_found = True
                answer = key

    return answers_found, answer


@inlineCallbacks
def smart_questions(
    session,
    question,
    answer_dictionary,
    question_try_again=None,
    waiting_time=5,
    number_attempts=3,
    debug=False,
):
    """
    This function asks a question and waits for the user to respond.
    It compares the user response with the answer dictionary and returns the answer.
    The robot will ask the question again if the user response is not clear, up to 3 times.
    The robot will wait for 5 seconds for the user response.
    Args:
        session (Component): The session object.
        question (str): The question to be asked.
        answer_dictionary (dict): A dictionary with the answers.
        question_try_again (list): A list of questions to be asked again if the user response is not clear.
        waiting_time (int): The time to wait for the user response in seconds.
        number_attempts (int): The number of attempts to ask the question again.
        debug (bool): A flag to print debug information.

    Returns:
        str: The answer found in the user response.
    """
    if question_try_again is None:
        question_try_again = [
            "Sorry, can you repeat the answer?",
            "I couldn't hear the answer, can you repeat it again?",
            "I am not sure I can hear you, can you repeat?",
        ]
    # Check if the arguments are of the correct type
    if not isinstance(question, str):
        raise TypeError("question is not a string")
    if not isinstance(answer_dictionary, dict):
        raise TypeError("answer_dictionary is not a dictionary")
    # check if the dictionary contains only strings
    else:
        for key in answer_dictionary.keys():
            if not isinstance(key, str):
                raise TypeError("answer_dictionary is not a dictionary of strings")
            if not isinstance(answer_dictionary[key], list):
                raise TypeError("answer_dictionary is not a dictionary of lists")
            for value in answer_dictionary[key]:
                if not isinstance(value, str):
                    raise TypeError(
                        "answer_dictionary is not a dictionary of lists of strings"
                    )
    if question_try_again is not None and not isinstance(question_try_again, list):
        raise TypeError("question_try_again is not a list")
    # check if the list contains only strings
    else:
        for questionz in question_try_again:
            if not isinstance(questionz, str):
                raise TypeError("question_try_again is not a list of strings")
    if not isinstance(waiting_time, int):
        raise TypeError("waiting_time is not an integer")
    if not isinstance(number_attempts, int):
        raise TypeError("number_attempts is not an integer")
    if not isinstance(debug, bool):
        raise TypeError("debug is not a boolean")

    timer = 0
    attempt = 0

    yield session.call("rie.dialogue.say", text=question)

    # subscribes the asr function with the input stt stream
    yield session.subscribe(listen_smart_question, "rie.dialogue.stt.stream")
    # calls the stream. From here, the robot prints each 'final' sentence
    yield session.call("rie.dialogue.stt.stream")

    if user_response != "":
        print("User response: ", user_response)

    # loop while user did not say goodbye or bye

    while True:
        found_answer, answer = find_the_answer(answer_dictionary)
        if found_answer:
            yield session.call("rie.dialogue.stt.close")
            return answer

        timer += 0.5
        yield sleep(0.5)
        if timer >= waiting_time:
            attempt += 1
            if attempt >= number_attempts:
                yield session.call("rie.dialogue.stt.close")
                return answer
            else:
                timer = 0
                yield session.call(
                    "rie.dialogue.say", text=question_try_again[randint(0, 2)]
                )



import base64
import cv2
import numpy as np


def show_camera_stream(frame):
    # check if the frame is not empty
    if frame is None:
        raise ValueError("The frame is empty")
    # check if the frame is a dictionary
    if not isinstance(frame, dict):
        raise TypeError("The frame is not a dictionary")

    frame_single = frame["data"]["body.head.eyes"]
    # make sure the frame is byte-like and not a string; it's in base64
    frame_single = bytes(frame_single, "utf-8")
    # Decode the base64 string
    image_data = base64.b64decode(frame_single)

    # Convert the decoded bytes to a numpy array
    nparr = np.frombuffer(image_data, np.uint8)

    # Decode the numpy array into an image
    image = cv2.imdecode(nparr, cv2.IMREAD_COLOR)

    # Display the image and close it in 1 second
    cv2.imshow("Camera Stream", image)
    cv2.waitKey(100)



import cv2
import base64
import numpy as np
from twisted.internet.defer import inlineCallbacks


def detect_face_in_frame(frame):
    """Function to detect a face in a frame using OpenCV's Haar Cascade classifier

    Args:
        frame (dictionary):
        The frame dictionary from the robot's camera stream

    Returns:
        tuple: (top_left, bottom_right)
        The coordinates of the detected face in the frame

    """
    # Extract and preprocess the frame
    frame_single = frame["data"]["body.head.eyes"]
    # Ensure the frame is byte-like and decode from base64
    frame_single = bytes(frame_single, "utf-8")
    image_data = base64.b64decode(frame_single)

    # Convert the byte data to a NumPy array and decode it into an image
    np_array = np.frombuffer(image_data, dtype=np.uint8)
    image = cv2.imdecode(np_array, cv2.IMREAD_COLOR)

    # Convert to grayscale for face detection
    gray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)

    # Load OpenCV's pre-trained Haar Cascade for face detection
    face_cascade = cv2.CascadeClassifier(
        cv2.data.haarcascades + "haarcascade_frontalface_default.xml"
    )

    # Detect faces in the image
    faces = face_cascade.detectMultiScale(
        gray, scaleFactor=1.1, minNeighbors=5, minSize=(30, 30)
    )

    # If no face is detected, return None
    if len(faces) == 0:
        return None

    # Return the coordinates of the first detected face (x, y, width, height)
    x, y, w, h = faces[0]
    top_left = (x, y)
    bottom_right = (x + w, y + h)
    if top_left is not None:
        return top_left, bottom_right


@inlineCallbacks
def center_face(session, frame):
    center = None
    result = detect_face_in_frame(frame)
    if result is not None:
        top_left, bottom_right = result
        # Calculate the center of the detected face
        center = (
            (top_left[0] + bottom_right[0]) // 2,
            (top_left[1] + bottom_right[1]) // 2,
        )

    if center and center[0] > 155:
        motors = yield session.call("rom.sensor.proprio.read")
        head_motors = motors[0]["data"]["body.head.yaw"]
        frames = [
            {
                "time": 100,
                "data": {
                    "body.head.yaw": head_motors - 0.06,
                },
            },
        ]
        yield session.call(
            "rom.actuator.motor.write", frames=frames, force=True, sync=True
        )
    elif center and center[0] < 125:
        motors = yield session.call("rom.sensor.proprio.read")
        head_motors = motors[0]["data"]["body.head.yaw"]
        frames = [
            {
                "time": 100,
                "data": {
                    "body.head.yaw": head_motors + 0.06,
                },
            },
        ]
        yield session.call(
            "rom.actuator.motor.write", frames=frames, force=True, sync=True
        )


@inlineCallbacks
def follow_face(session):
    """Function to subscribe to the robot's camera stream and follow a detected face

    Args:
        session (Component):
        The session object for the connection to the robot

    """

    # Wrapper function to pass the session
    def center_face_wrapper(frame):
        return center_face(session, frame)

    # Subscribe to the camera stream
    yield session.subscribe(center_face_wrapper, "rom.sensor.sight.stream")
    yield session.call("rom.sensor.sight.stream")
    pass



from autobahn.twisted.component import Component, run
from twisted.internet.defer import inlineCallbacks


# dictionary of joint angles
# joint name: (min_angle, max_angle, minimum time from 0 to min or max)":
joints_dic = {
    "body.head.yaw": (-0.874, 0.874, 600),  # 0.3
    "body.head.roll": (-0.174, 0.174, 400),  # 0.2
    "body.head.pitch": (-0.174, 0.174, 400),  # 0.2
    "body.arms.right.upper.pitch": (-2.59, 1.59, 1800),  # 0.8
    "body.arms.right.lower.roll": (-1.74, 0.000064, 1200),  # 0.6
    "body.arms.left.upper.pitch": (-2.59, 1.59, 1600),  # 0.8
    "body.arms.left.lower.roll": (-1.74, 0.000064, 1200),  # 0.6
    "body.torso.yaw": (-0.874, 0.874, 1000),  # 0.5
    "body.legs.right.upper.pitch": (
        -1.74,
        1.74,
        1600,
    ),  # 0.8 + add warning about falling somehow
    "body.legs.right.lower.pitch": (
        -1.74,
        1.74,
        1200,
    ),  # 0.6 + add warning about falling somehow
    "body.legs.right.foot.roll": (
        -0.849,
        0.249,
        1000,
    ),  # 0.5 + add warning about falling somehow
    "body.legs.left.upper.pitch": (
        -1.74,
        1.74,
        1600,
    ),  # 0.8 + add warning about falling somehow
    "body.legs.left.lower.pitch": (
        -1.74,
        1.74,
        1200,
    ),  # 0.6 + add warning about falling somehow
    "body.legs.left.foot.roll": (
        -0.849,
        0.249,
        1000,
    ),  # 0.5 + add warning about falling somehow
}


def check_angle_set_value(frame_joints_dic):
    for joint in frame_joints_dic:
        if not joint in joints_dic:
            raise ValueError(joint + " is not a valid joint name")
        else:
            if (
                not joints_dic[joint][0]
                <= frame_joints_dic[joint]
                <= joints_dic[joint][1]
            ):
                raise ValueError(
                    "The angle selected for joint " + joint + " is out of bounds"
                )


def calculate_required_time(current_pos, target_pos, min_angle, max_angle, min_time):
    # calculate the total range of motion
    total_range = abs(max_angle - min_angle)

    # calculate the movement range required
    movement_range = abs(target_pos - current_pos)

    # calculate the proportional time
    proportional_time = (movement_range / total_range) * min_time

    return proportional_time


# the minimum time between any movements is proportional to the angle of movement
@inlineCallbacks
def perform_movement(
    session, frames, mode="linear", sync=True, force=False
):
    """
    This function performs a movement with the robot's joints. The time of each frame is calculated based on the proportional time of the movement.

    Args:
                session (Component): The session object.
                frames (list): A list of dictionaries with the time and data of the joints to be moved.
                mode (str): The mode of the movement. Choose one of the following: "linear", "last", "none".
                sync (bool): A flag to synchronize the movement of the joints.
                force (bool): A flag to force the movement of the joints.

        Returns:
                None
    """
    if not isinstance(frames, list) and all(isinstance(item, dict) for item in frames):
        raise TypeError(
            'frames is not a list of tuples, it needs to follow the structure [{"time": (int), "data": {name_joints (string): position_joint (float), ...}}'
        )

    if not isinstance(mode, str):
        raise TypeError(
            'mode is not a string, choose one of the following "linear", "last", "none"'
        )

    if not isinstance(sync, bool):
        raise TypeError(
            "sync is not a boolean, choose one of the following True, False"
        )

    if not isinstance(force, bool):
        raise TypeError(
            "force is not a boolean, choose one of the following True, False"
        )

        # check the joints and angles of the first frame
    check_angle_set_value(frames[0]["data"])

    # get the joints angle at this time
    current_position = yield session.call("rom.sensor.proprio.read")
    # print(current_position[0]["data"])

    # check the time set of the first frame
    for joint, target_position in frames[0]["data"].items():
        minimum_required_time = calculate_required_time(
            current_position[0]["data"][joint],
            target_position,
            joints_dic[joint][0],
            joints_dic[joint][1],
            joints_dic[joint][2],
        )

        minimum_required_time = round(minimum_required_time, 2)
        print(minimum_required_time)
        if frames[0]["time"] == None or minimum_required_time > frames[0]["time"]:
            print(
                "The time of frame 0 was changed from "
                + str(frames[0]["time"])
                + " to "
                + str(minimum_required_time)
            )
            frames[0]["time"] = minimum_required_time

    for idx in range(len(frames) - 1):
        frame1 = frames[idx]
        frame2 = frames[idx + 1]

        for joint, target_position in frame2["data"].items():
            minimum_required_time = calculate_required_time(
                frame1["data"][joint],
                target_position,
                joints_dic[joint][0],
                joints_dic[joint][1],
                joints_dic[joint][2],
            )
            minimum_required_time = round(minimum_required_time, 2)
        if frame2["time"] == None or minimum_required_time > frame2["time"]:
            print(
                "The time of frame "
                + str(idx + 1)
                + " was changed from "
                + str(frame2["time"])
                + " to "
                + str(minimum_required_time)
            )
            frame2["time"] = minimum_required_time

    session.call(
        "rom.actuator.motor.write", frames=frames, mode=mode, sync=sync, force=True
    )



from setuptools import setup, find_packages

setup(
    name="alpha_mini_rug",
    version="0.1.7",
    description="Alpha Mini Robot wrapper for the RUG Social Robotics Lab",
    author="RUG Social Robotics Lab",
    packages=find_packages(),
    install_requires=[
        "txaio",
        "autobahn",
        "numpy",
        "opencv-python",
        "opencv-contrib-python",
    ],
)



from .aruco import aruco_detect_markers
from .smart_q import smart_questions
from .keywords import key_words
from .follow_face import follow_face
from .movements import perform_movement
from .camera_services import show_camera_stream
